import numpy as numpy
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt
import matplotlib.pyplot as plt

from .SplinesModel import SplinesModel
from .PolynomialModel import PolynomialModel


from .kernels.Kernel import Kernel
from .kernels.Tophat import Tophat

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2020 - 2021 Do Kester

class BasicSplinesModel( SplinesModel ):
    """
    Splines model consisting of a basis of spline blobs.

    The blobs have limited support. Each blob is a segment of polynomial order,
    between 2 knots. At the knots they are continuous (differentiable) upto order - 1.
    Similarly the edges of the blobs are smoothly connected to 0.

    order   support behaviour between knots     continuity at knots
      0        1    piecewise constant          not continuous at all
      1        2    piecewise linear            lines are continuous (connected)
      2        3    parabolic pieces            1st derivatives are also continuous
      3        4    cubic pieces                2nd derivatives are also continuous
     n>3      n+1   n-th order polynomials      (n-1)-th derivatives are also continuous

    The function result is the sum over all spline blobs, multiplied with
    the parameters, the amplitudes of the spline blobs.

    The support of the knots defined the domain where the function is defined. They are
    hard edges. Consequently the function is not continuous or differentiable at the edges.
    The spline blobs at the edges may be different from the ones in the middle.


    From SplinesModel
    -----------------
    The user lays out a number ( << datapoints ) of knots on the x-axis at
    arbitrary position, generally more knots where the curvature is higher.
    The knots need to be monotonuously increasing in x.
    Alternatively one can ask this class to do the lay-out which is then
    equidistant in x over the user-provided range.
    Through these knots a splines function is obtained which best
    fits the datapoints. One needs at least 2 knots, one smaller and one
    larger than the x-values in the dataset.

    This model is NOT for (cubic) spline interpolation.

    Examples
    --------
    >>> knots = numpy.arange( 17, dtype=float ) * 10    # make equidistant knots from 0 to 160
    >>> csm = BasicSplinesModel( knots=knots, order=2 )
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> csm = SplinesModel( nrknots=17, order=2, min=0, max=160 )    # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18
    # or alternatively:
    >>> npt = 161                                               # to include both 0 and 160.
    >>> x = numpy.arange( npt, dtype=float )                    # x-values
    >>> csm = BasicSplinesModel( nrknots=17, order=2, xrange=x )     # automatic layout of knots
    >>> print csm.getNumberOfParameters( )
    18

    Attributes
    ----------
    knots : array_like
        positions of the spline knots
    order : int
        order of the spline. default: 3

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    Limitations
    -----------
    Dont construct the knots so closely spaced, that there are no datapoints in between.

    """
    def __init__( self, knots=None, order=3, nrknots=None, min=None, max=None, xrange=None,
                        border=0, copy=None, **kwargs ):
        """
        Splines on a given set of knots and a given order.

        The number of parameters is ( length( knots ) + order - 1 )

        Parameters
        ----------
        knots : array_like
            a array of arbitrarily positioned knots
        order : int
            order of the spline. Default 3 (cubic splines)
        nrknots : int
            number of knots, equidistantly posited over xrange or [min,max]
        border : [0, 1, 2]
            defines what happens at the borders of the knot range.
            0 : Just like de Boors b-splines.
                the model is NOT defined outside the knot range.
            1 : periodic, make knot[0] the same as knot[-1]
            2 : easy borders. the model is slightly extensable.
        min : float
            minimum of the knot range
        max : float
            maximum of the knot range
        xrange : array_like
            range of the xdata
        copy : BasicSplinesModel
            model to be copied.
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        Raises
        ------
        ValueError : At least either (`knots`) or (`nrknots`, `min`, `max`) or
                (`nrknots`, `xrange`) must be provided to define a valid model.

        Notes
        -----
        The SplinesModel is only strictly valid inside the domain defined by the
        minmax of knots. It deteriorates fastly going outside the domain.

        """
        if knots is None and nrknots is None :
            raise ValueError( "Need either knots or (nrknots,min,max) or (nrknots,xrange)" )

        npar = nrknots if knots is None else len( knots )
        npar += order - 1 if border == 0 else -1

        super( ).__init__( nparams=npar, knots=knots, order=order, nrknots=nrknots,
                           min=min, max=max, xrange=xrange, copy=copy, **kwargs )

        self.border = border
        self.makeBasis = self.makeBaseBasis if border == 0 else self.makePeriodicBasis

        if copy is None :
            self.poly = PolynomialModel( self.order )
            self.basis = self.makeBasis()
        else :
            self.poly = copy.poly
            self.basis = copy.basis.copy()



    def copy( self ):
        return BasicSplinesModel( knots=self.knots, border=self.border, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: knots, order

        """
        if name == "basis" :
            setatt( self, name, value )
        elif name == "border" :
            setatt( self, name, value, type=int )
        elif name == "poly" :
            setatt( self, name, value, type=PolynomialModel )
        elif name == "makeBasis" :
            setatt( self, name, value )
        else :
            super( ).__setattr__( name, value )

    def makeBaseBasis( self ) :
        """
        Make a sets of polynomial bases for each of the parameters

        Return
        ------
        basis : 3-d array-like
            parameters to the polynomials that make up the spline blobs

        """
        np = self.order + 1
        nk = len( self.knots )
        hi = 1
        lo = 0

        basis = numpy.zeros( (nk-1,np,self.npmax), dtype=float )
        for k in range( self.npmax ) :

            if hi < nk : hi += 1
            if k >= np : lo += 1

            knotix = numpy.arange( lo, hi, dtype=int )
            dist = self.makeDist( knotix )

            bpar = self.findParameters( knotix, dist, kpar=k )
            bpar = numpy.reshape( bpar, ( -1, np ) )

            basis[knotix[:-1],:,k] =  bpar

        basis = self.normalizeBasis( basis )

        return basis

    def makeDist( self, knotix ) :
        return self.knots[knotix[:-1]+1] - self.knots[knotix[:-1]]


    def makePeriodicBasis( self ) :
        """
        Make a sets of polynomial bases for each of the parameters

        Return
        ------
        basis : 3-d array-like
            parameters to the polynomials that make up the spline blobs

        """
        if self.order == 0 :
            return self.makeBaseBasis()

        np = self.order + 1
        nk = len( self.knots )
        nh = ( np + 1 ) // 2

        basis = numpy.zeros( (nk-1,np,self.npmax), dtype=float )

        knotix = numpy.zeros( np + 1, dtype=int )
        knotix[:np] = numpy.arange( np, dtype=int ) - np - 1

        for k in range( self.npmax ) :
            dist = self.knots[knotix+1] - self.knots[knotix]

            bpar = self.findParameters( knotix, dist )
            bpar = numpy.reshape( bpar, ( -1, np ) )


            km = knotix[:-1] % nk
            basis[km,:,k] =  bpar

            knotix = numpy.where( knotix == -2, 0, knotix + 1 )

        basis = self.normalizeBasis( basis )
        return basis

    def normalizeBasis( self, basis ) :
        """
        Normalize the base splines such that a constant value of 1.0
        is returned when all model parameters are 1.

        Parameters
        ----------
        basis : array_like
            parameters to the polynomials that make up the spline blobs
        """

        ## normalize the blobs to 1
        np = self.npmax
        nk = len( self.knots )
        mat = numpy.zeros( (np,np), dtype=float )

        if self.order % 2 == 0 :
            xmid = ( self.knots[1:] + self.knots[:-1] ) / 2
        else :
            k = 1 - self.border             ## only for border 0 and 1
            xmid = self.knots[k:-1]

        if self.border == 1 or self.order == 0 :
            xx = xmid
        elif nk == 2 :
            xx = numpy.linspace( self.knots[0], self.knots[-1], np, dtype=float )
        else :
            xx = numpy.zeros( np, dtype=float )
            xx[0] = self.knots[0]
            xx[-1] = self.knots[-1]

            ne = np - nk
            ds = ( self.knots[1]  - self.knots[0]  ) / ne
            de = ( self.knots[-1] - self.knots[-2] ) / ne
            ks = self.knots[0] + ds
            ke = self.knots[-1] - de
            ne = ne // 2
            for k in range( ne ) :
                xx[ 1+k] = ks
                xx[-2-k] = ke
                ks += ds
                ke -= de

            xx[1+ne:-1-ne] = xmid

        x2k = self.makeKnotIndices( xx )

        for k in range( np ) :
            mat[:,k] = self.basicBlob( xx, basis[:,:,k], x2k, self.poly )

        beta = numpy.ones( np, dtype=float )
        bpar = numpy.linalg.solve( mat, beta )

        return basis * bpar


    def findParameters( self, knotix, dist, kpar=0 ) :
        """
        Find the parameters by assuming (order-1) continuous differentials.
        At the edges it is less. Normalized to 1.0


        Parameters
        ----------
        knotix : int array
            knot indices involved in this spline blob
        dist : array_like
            distances between knots
        kpar : int
            index of parameter for which the spline-blob is constructed

        Returns
        -------
        par : 2-d array
            sets of poly parameters.
        """
        lo = knotix[0]
        hi = knotix[-1] + 1

        knots = self.knots
        order = self.order
        k0 = order + 1
        np = ( len( knotix ) - 1 ) * k0
        nk = len( knots )

        mat = numpy.zeros( ( np, np ), dtype=float )
        kdf = numpy.arange( k0, dtype=int )

        i = 0
        n = 0
        ks = 0

        ## handling of the left side (edge)
        ## values and diffs all zero
        ff = numpy.ones( k0, dtype=float )
        x0 = numpy.power( 0.0, kdf )                ## center (== 0 ) is at each left knot

        ne1 = min( order, kpar ) if self.border == 0 else order
        while n < ne1 :
            mat[n, kdf] = x0 * ff
            x0 = numpy.roll( x0, 1 )
            ff *= numpy.roll( kdf, n )
            n += 1

        ## for normalization
        if n == 0 :
            mat[-1,i+kdf] = x0

        ## non-edge knots : both poly pieces (and diffs) are same
        ne = min( ks + k0, hi - 1 )
        ne = ks + len( knotix ) - 2

        while ks < ne :
            x0 = numpy.power( 0.0, kdf )    ## value of the left side knot
            kc = dist[ks]
            xx = numpy.power( kc, kdf )     ## value of the right side knot

            # normalization
            if 0 <= i < np : mat[-1,i+kdf] = xx

            ## values at xx (left side params) are equal to values at x0 (right side params)
            ff = numpy.ones( k0, dtype=float )
            idf1 = i + kdf
            idf2 = idf1 + k0
            for k in range( order ) :
                mat[n, idf1] = +xx * ff
                mat[n, idf2] = -x0 * ff
                xx = numpy.roll( xx, 1 )
                x0 = numpy.roll( x0, 1 )
                ff *= numpy.roll( kdf, k )
                n += 1

            i += k0
            ks += 1

        ## right side edge
        ff = numpy.ones( k0, dtype=float )
        xx = numpy.power( dist[ks], kdf )

        k = 0
        while n < np - 1 :
            mat[n, i+kdf] = xx * ff
            xx = numpy.roll( xx, 1 )
            ff *= numpy.roll( kdf, k )
            n += 1
            k += 1

        """
        ## more normalization
        if kpar > 0 :
            mat[-1,-1] = 1.0             ## at right normalize to 1.0

        """
        if kpar > 0 :
            if len( knotix ) == 2 :
                cc = ( ( np - kpar ) * knots[0] + kpar * knots[1] ) / np
                mat[-1,i+kdf] = numpy.power( cc, kdf )
            else :
                mat[-1,i] = 1.0             ## at center normalize to 1.0

        ## beta vector: all zero except last one, is 1 (normalization)
        beta = numpy.zeros( np, dtype=float )
        beta[-1] = 1.0

#        print( fmt( mat, max=None ) )

        try :
            ## solve linear set of equations
            bspar = numpy.linalg.solve( mat, beta )
        except numpy.linalg.LinAlgError :
            mat[-1,-1] = 1.0
            bspar = numpy.linalg.solve( mat, beta )

        return bspar


    def baseResult( self, xdata, params ):
        """
        Returns the functional result at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)

        """
        np = self.npmax
        na = self.order + 1

        x2k = self.makeKnotIndices( xdata )

        pa = numpy.inner( self.basis, params )
        res = numpy.zeros_like( xdata )
        for k, kn in enumerate( self.knots[:-1] ) :
            q = numpy.where( x2k == k )
            xc = xdata[q] - kn
            res[q] = self.poly.result( xc, pa[k,:] )

        return res


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        The partials are the powers of x (input) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model (ignored in LinearModels)
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        np = self.npbase
        ni = Tools.length( xdata )
        partial = numpy.zeros( ( ni, np), dtype=float )

        if parlist is None :
            parlist = range( self.npmax )

        x2k = self.makeKnotIndices( xdata )
        for k,kb in enumerate( parlist ) :
            bss = self.basis[:,:,kb]
            partial[:,k] = self.basicBlob( xdata, bss, x2k, self.poly )

        return partial

        """
        for kb in range( np ) :
            bss = self.basis[:,:,kb]
            partial[:,kb] = self.basicBlob( xdata, bss, x2k, self.poly )

        return partial
        """

    def makeKnotIndices( self, xdata ) :
        """
        Return a list of indices of the knots immediately preceeding the xdata.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the indices
        """
        ix = numpy.zeros( len( xdata ), dtype=int )
        for kn in self.knots[1:-1] :
            ix = numpy.where( xdata >= kn, ix + 1, ix )

        return ix


    def basicBlob( self, xdata, basis, x2k, poly ) :
        """
        Calculates a spline blob for all of xdata

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the spline
        basis : array_like
            splineParameters
        x2k : int_array
            pointing to the knot preceeding each xdata point
        poly : PolynomialModel
            model to calculate the splines
        """
        blob = numpy.zeros_like( xdata )

        kn = self.knots[:-1]
        for k, kn in enumerate( self.knots[:-1] ) :
            if all( basis[k,:] == 0 ) : continue
            q = numpy.where( x2k == k )
            xc = xdata[q] - kn
            blob[q] = poly.result( xc, basis[k,:] )

        return blob

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters to the model

        """
        dfdx = numpy.zeros_like( xdata )
        if self.order == 0 :
            return dfdx

        polym1 = PolynomialModel( self.order - 1 )
        x2k = self.makeKnotIndices( xdata )

        fp = numpy.arange( self.order + 1 )
        for n in range( self.npmax ) :
            dbs = self.basis[:,:,n] * fp
            dbs = dbs[:,1:]
            dfdx += self.basicBlob( xdata, dbs, x2k, polym1 ) * params[n]

        return dfdx

    def baseName( self ):
        """ Returns a string representation of the model. """
        return "BasicSplines of order %d with %d knots."%( self.order, len( self.knots) )

    def baseParameterUnit( self, k ):
        """
        Return the name of the parameter.

        Parameters
        ----------
        k : int
            index of the parameter.
        """
        return self.yUnit




import numpy as numpy

from .Problem import Problem
from .Formatter import formatter as fmt

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2021 Do Kester

class MultipleOutputProblem( Problem ):
    """
    A MultipleOutputProblem is an optimization of parameters where the model
    has multiple outputs. E.g. the orbit of a double star or the outcome of
    a game.

    Problems can be solved by NestedSampler, with appropriate Engines and
    ErrorDistributions.

    The result of the function for certain x and p is given by
    problem.result( p )
    The parameters, p, are to be optimized while the x provide additional
    information.

    Attributes from Problem
    -----------------------
    model, xdata, ydata, weights, partype

    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, model=None, xdata=None, ydata=None, weights=None, copy=None ):
        """
        Problem Constructor.

        Parameters
        ----------
        model : Model
            the model to be solved. One with multiple outputs: model.ndout > 1
        xdata : array_like
            independent variable
        ydata : array_like
            dependent variable. shape = (len(xdata), model.ndout)
        weights : array_like or None
            weights associated with ydata: shape = as xdata or as ydata
        copy : Problem
            to be copied

        """
        if weights is not None :
            if len( weights.shape ) == 1 :
                wgts = numpy.zeros( ( len( weights ), model.ndout ), dtype=float )
                for k in range( model.ndout ) :
                    wgts[:,k] = weights
                weights = wgts
            weights = weights.flatten()

        super( ).__init__( model=model, xdata=xdata, ydata=ydata, weights=weights,
                           copy=copy )


    def copy( self ):
        """
        Copy.

        """
        return MultipleOutputProblem( copy=self )


    #  *****RESULT**************************************************************
    def result( self, param ):
        """
        Returns the result calculated at the xdata.

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        """
        return self.model.result( self.xdata, param )

    def partial( self, param ) :
        """
        Returns the partials (df/dp) calculated at the xdata.

        Parameters
        ----------
        param : array_like
            values for the parameters + nuisance params.

        """

        parts = self.model.partial( self.xdata, param )

        partial = parts[0]
        for k in range( 1, self.model.ndout ) :
            partial = numpy.append( partial, parts[k], 1 )

        return partial.reshape( -1, self.npars )


    def derivative( self, param ) :
        """
        Return the derivative of the internal model.

        Parameters
        ----------
        param : array_like
            list of model parameters
        """
        return self.model.derivative( self.xdata, param )

    def residuals( self, param, mockdata=None ) :
        """
        Returns residuals in a flattened array.
        """
        res = super().residuals( param, mockdata=mockdata )
        return res.flatten()

    def myEngines( self ) :
        """
        Return a default list of preferred engines
        """
        return ["galilean", "chord"]

    def myStartEngine( self ) :
        """
        Return the default preferred startengines
        """
        return "start"

    def myDistribution( self ) :
        """
        Return the name of the preferred error distribution
        """
        return "gauss"

    #  *****TOSTRING***********************************************************
    def __str__( self ):
        """ Returns a string representation of the model.  """
        return "MultipleOutputProblem of " + self.model






import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .OrderEngine import OrderEngine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester

class LoopEngine( OrderEngine ):
    """
    The LoopEngine tries to unloop a crossing loop.

    Only for 2 dimensional TS problems.

    Input order :
         0  1  2  3
        15 14  5  4
         7  6 13 12
         8  9 10 11

    The loop crosses between (5,6) and (13,14). By switching the positions
    of 6 and 13, and reversing the loop in between, a better solution is
    reached (triangle inequality)

    output order:
         0  1  2  3
        15 14  5  4
        12 13  6  7
        11 10  9  8

    This is NOT a random engine as it only steps in the uphill direction.

    It belongs to the class of generalized travelling salesman problems
    where the parameters of the problem is an ordered list.

    The walker is kept when the logLikelihood > lowLhood

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : LoopEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return LoopEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "LoopEngine" )

    #  *********EXECUTE***************************************************
    def executeOnce( self, kw, lowLhood, dims=[0,1] ) :
        """
        Execute the LoopEngine one time.

        Parameters
        ----------
        kw : int
            walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        dims : list of 2 ints
            dimensions to process over
        """
        walker = self.walkers[kw]

        problem = walker.problem
        np = problem.npars
        param = walker.allpars

        ## roll the parameter list to an arbitrary point
        ks = self.rng.randint( np )
        ptry = numpy.roll( param, ks )

        ## select last point and first point
        k1 = ptry[-1]
        k2 = ptry[0]
        x1, y1 = tuple( problem.xdata[k1,dims] )
        x2, y2 = tuple( problem.xdata[k2,dims] )
        if x1 == x2 :
            dims = [dims[1], dims[0]]           ## reverse dimensional order
            m1 = 0                              ## slope of point k1 to k2
        else :
            m1 = ( y2 - y1 ) / ( x2 - x1 )

        k4 = ptry[1]
        for t in range( 2, np - 1 ) :       ## skip adjacent points
            k3 = k4
            k4 = ptry[t]
            x3, y3 = tuple( problem.xdata[k3,dims] )
            x4, y4 = tuple( problem.xdata[k4,dims] )

            if x3 == x4 :
                xc = x3
                yc = y1 + m1 * ( xc - x1 )
                if Tools.isBetween( x1, xc, x2 ) and Tools.isBetween( y3, yc, y4 ) :
                    break

            else :
                m3 = ( y4 - y3 ) / ( x4 - x3 )
                xc = ( y3 - y1 + m3 * x3 - m1 * x1 ) / ( m3 - m1 )
                if Tools.isBetween( x1, xc, x2 ) and Tools.isBetween( x3, xc, x4 ) :
                    break

        else :
            self.reportReject()
            return 0


        ## reverse first t items
        ptry[:t] = ptry[t-1::-1]

        ptry = numpy.roll( ptry, -ks )

        Ltry = self.errdis.logLikelihood( problem, ptry )

        if self.verbose > 4 :
            print( "Loop     ", ks, t, fmt( lowLhood ), fmt( Ltry ) )
            print( fmt( param, max=None, format='%3d' ) )
            print( fmt( ptry, max=None, format='%3d' ) )

        if Ltry >= lowLhood:
            self.reportSuccess( )
            self.setWalker( walker.id, problem, ptry, Ltry )

        else :
            self.reportFailed()
            t = 0

        return t                        # nr of succesfull steps





from .KernelModel import KernelModel
from .kernels.Sinc import Sinc

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class SincModel( KernelModel ):
    """
    Sinc Model.
    Also known as Cardinal Sine.

        f( x:p ) = p_0 * sin( ( x - p_1 ) / p_2 ) / ( ( x - p_1 ) / p_2 )

    where
        p_0 = amplitude
        p_1 = offset
        p_2 = width ( =Distance between first zero-crossings divided by 2 Pi. )
    As always x = input.

    The parameters are initialized at {1.0, 0.0, 1.0}.
    Parameter 2 ( width ) is always kept positive ( >=0 ).

    SincModel() is syntactic sugar for
        KernelModel( kernel=Sinc() )
    See @KernelModel

    Examples
    --------
    >>> sinc = SincModel( )
    >>> print sinc.npchain
    3
    >>> print( sinc( numpy.arange( 15, dtype=float )-7 ) )        # sinc function between [-7,+7]
    [0,0,0,0,0,0,0,1,0,0,0,0,0,0,0]


    Attributes
    ----------
    none of its own

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """
    def __init__( self, copy=None, **kwargs ):
        """
        Sinc model.

        Number of parameters is 3.

        Parameters
        ----------
        copy : ArctanModel
            to be copied
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
        """
        super( SincModel, self ).__init__( kernel=Sinc(), copy=copy, **kwargs )

    def copy( self ):
        """ Copy method.  """
        return SincModel( copy=self )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .NonLinearModel import NonLinearModel
from .Kepplers2ndLaw import Kepplers2ndLaw

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2020 Do Kester

class RadialVelocityModel( NonLinearModel ) :
    """
    Model for the radial velocity variations of a star caused by a orbiting planet.

    p_0 : eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
    p_1 : amplitude    of the velocity variation (>0)
    p_2 : period       of the velocity variation (>0)
    p_3 : phase        phase of periastron (0<p<2pi)
    p_4 : periastron   longitude of periastron (0<p<2pi)

    This class uses @Kepplers2ndLaw to find radius and true anomaly.

    Note:
    The velocity of the star system is not included in this model. See example.

    The parameters are initialized at {0.0, 1.0, 1.0, 0.0, 0.0}.
    It is a non-linear model.

    Attributes
    ----------
    keppler : Kepplers2ndLaw()
        to calculate the radius and true anomaly

    Examples
    --------
    >>> rv = RadialVelocityModel( )
    >>> print( rv.npars )
    5
    >>> rv += PolynomialModel( 0 )          # add a constant system velocity

    """
    TWOPI = 2 * math.pi

    def __init__( self, copy=None, **kwargs ):
        """
        Radial velocity model.

        Number of parameters is 5:

        Parameters
        ----------
        copy : RadialVelocityModel
            model to copy
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        param = [0.0, 1.0, 1.0, 0.0, 0.0]
        names = ["eccentricity", "amplitude", "period", "phase", "longitude"]

        super( RadialVelocityModel, self ).__init__( 5, copy=copy, params=param,
                        names=names, **kwargs )

        setatt( self, "keppler", Kepplers2ndLaw() )

    def copy( self ):
        """ Copy method.  """
        return RadialVelocityModel( copy=self )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == "semimajoraxis" :
            p = self.parameters
            return p[0] * p[1] * math.sqrt ( 1 - p[2] * p[2] ) / self.TWOPI
        else :
            return super( RadialVelocityModel, self ).__getattr__( name )

    def getMsini( self, stellarmass ) :
        """
        Return the mass of the exoplanet in Jupiter masses.

        Parameters
        ----------
        stellarmass : float
            mass of the host star in solar masses.
        """
        p = self.parameters
        return ( 4.91e-3 * math.pow( stellarmass, 2.0 / 3.0 ) * p[0] *
                 math.pow( p[1], 1.0 / 3.0 ) * math.sqrt( 1 - p[2] * p[2] ) )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        f(x:p) = p_1 * ( cos( v + p_4 ) + p_0 * cos( p_4 ) )

        where v is the true anomaly

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        pars = params[:4]
        r, v = self.keppler.radiusAndTrueAnomaly( xdata, pars )

        x = v + params[4]

        result = params[1] * ( numpy.cos( x ) + params[0] * math.cos( params[4] ) )
        return result


    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        f(x:p) = p_1 * ( cos( v + p_4 ) + p_0 * cos( p_4 ) )

        df/dp_0 = p_1 * ( - sin( v + p_4 ) dv/dp_0 + cos( p_4 ) )
        df/dp_1 = cos( v + p_4 ) + p_0 * cos( p_4 )
        df/dp_2 = - p_1 * sin( v + p_4 ) dv/dp_2
        df/dp_3 = - p_1 * sin( v + p_4 ) dv/dp_3
        df/dp_4 = - p_1 * ( sin( v + p_4 ) + p_0 * sin( p_4 ) )

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        ecc = params[0]
        amp = params[1]
        lon = params[4]

        np = self.npbase if parlist is None else len( parlist )

        pars = params[:4]
        r, v = self.keppler.radiusAndTrueAnomaly( xdata, pars )

        E = self.keppler.eccAnomaly
        sinE = self.keppler.sinE
        cosE = self.keppler.cosE

        drde, drda, drdP, drdp, dvde, dvdP, dvdp = self.keppler.drvdpar( xdata, pars,
                                       E, cosE, sinE )

        vp4 = v + lon
        sinvp4 = numpy.sin( vp4 )


        part = { 0 : ( lambda : amp * ( math.cos( lon ) - sinvp4 * dvde ) ),
                 1 : ( lambda : numpy.cos( vp4 ) + ecc * math.cos( lon ) ),
                 2 : ( lambda : -amp * sinvp4 * dvdP ),
                 3 : ( lambda : -amp * sinvp4 * dvdp ),
                 4 : ( lambda : -amp * ( sinvp4 + ecc * math.sin( lon ) ) ) }

        partial = numpy.ndarray( ( Tools.length( xdata ), np ) )

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = part[kp]()


        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative of f to x (df/dx) at the input values.

        dfdx = - p_1 * sin( v + p_4 ) * dvdx

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        pars = params[:4]
        r, v = self.keppler.radiusAndTrueAnomaly( xdata, pars )

        sinE = self.keppler.sinE
        cosE = self.keppler.cosE
        drdx, dvdx = self.keppler.drvdx( xdata, pars, cosE, sinE )

        return - params[1] * numpy.sin( v + params[4] ) * dvdx

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "RadialVelocity " )

    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter.

        Parameters
        ----------
        k : int
            the kth parameter.

        """
        if k == 0:
            return units.Unit( units.si.rad ) / self.xUnit
        return self.yUnit




#run with: python3 -m unittest TestModifiable

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

#from StdTests import stdModeltest

from BayesicFitting import *

from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        self.dofull = ( "DOFULL" in os.environ and os.environ["DOFULL"] == "1" )

    def makeData( self, npt=51, noise=0.1 ) :
        t = numpy.linspace( 0, 100, npt, dtype=float )
#        ym = numpy.sin( 2 * math.pi * numpy.exp( t / 150 ) )
        ym = 2 * numpy.sin( 2 * math.pi * numpy.exp( t / 80 ) +1 ) * numpy.exp( -0.02 * t )
        y = numpy.random.seed( 12345 )
        y = ym + numpy.random.randn( npt ) * noise

        return t, y


    def test1( self ) :
        print( "1  Test Splines" )

        t, y = self.makeData()
        knots =[0, 20, 40, 60, 75, 90, 100]

        mdl = BasicSplinesModel( knots=knots )
        mdl.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        engs = ["galilean", "chord"]

        ns = NestedSampler( t, mdl, y, seed=1234, engines=engs )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2
        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        evid = ns.sample( plot=self.doplot )

    def test2( self ) :
        print( "2  Test Modifiable Splines" )

        t, y = self.makeData()
        knots =[0, 10, 25, 40, 50, 60, 75, 100]

        mdl = SplinesDynamicModel( knots=knots, dynamic=False )
        mdl.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        engs = ["galilean", "struct"]

        ns = NestedSampler( t, mdl, y, seed=31234, engines=engs )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2
#        ns.engines[1].slow = 5
        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        ka = numpy.zeros( ( len( knots ), len( sl ) ), dtype=float )
        for k,s in enumerate( sl ) :
            ka[:,k] = s.model.knots
        for j in range( len( knots ) ) :
            plt.plot( ka[j,:], cc[j % 6] )
        plt.show()

    def test3( self ) :
        print( "3  Test Dynamic Splines" )

        t, y = self.makeData()
        knots =[0, 100]

        mdl = SplinesDynamicModel( knots=knots )
        mdl.setLimits( lowLimits=[-5.0], highLimits=[+5.0] )

        engs = ["galilean", "birth", "death", "struct"]

        ns = NestedSampler( t, mdl, y, seed=1234, engines=engs )
        ns.distribution.setLimits( [0.01,1] )
        ns.verbose = 2
        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
#        ka = numpy.zeros( ( len( knots ), len( sl ) ), dtype=float )
        for k,s in enumerate( sl ) :
            kn = s.model.knots
            nn = [k] * len( kn )
            plt.plot( nn, kn, cc[2] )
        wgts = sl.getWeightEvolution()
        mw = max( wgts )
        plt.plot( 100 * wgts / mw, 'k-' )
        plt.show()

    def test4( self ) :
        print( "4  Test Modifiable Splines compound" )

        t, y = self.makeData()
        NP = 41
        t = numpy.linspace( 0, 100, NP, dtype=float )
        dt = GaussModel().result( t, [10.0, 34, 10] )

        ym = numpy.sin( 0.1 * ( t + dt ) )
        y = ym + 0.1 * numpy.random.randn( NP )


        knots =[0, 25, 50, 75, 100]
        xm = SplinesDynamicModel( knots=knots, dynamic=False )
        xm += PowerModel( 1, fixed={0:1.0} )
        xm.setPrior( 0, LaplacePrior( center=0, scale=4 ) )

        sm = SineModel()
        sm.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        mdl = xm | sm

        ns = NestedSampler( t, mdl, y, seed=31234 )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2
#        ns.engines[2].verbose = 5
        ns.engines[2].slow = 5
        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10

        evid = ns.sample( plot=self.doplot )

        return

        if not self.doplot : return

        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        yfit = numpy.zeros( NP, dtype=float )
        for s in sl :
            knts = s.model.knots
            yft = BasicSplinesModel( knts ).result( t, s.parameters[:7] )
            yfit += yft * s.weight
        plt.plot( t, yfit, 'b-' )
        plt.plot( t, ns.yfit, 'g-' )
        plt.plot( t, y, 'k.' )
        plt.show()


    def test5( self ) :
        print( "5  Test Dynamic Modifiable Splines: evidence with fixed scale " )

        t, y = self.makeData()

#        npt = 201
#        t = numpy.linspace( 0, 100, npt, dtype=float )
#        ym = 2 * numpy.sin( 2 * math.pi * numpy.exp( t / 80 ) +1 ) * numpy.exp( -0.02 * t )
#        y = numpy.random.seed( 12345 )
#        y = ym + numpy.random.randn( npt ) * 0.05

        knots =[0, 100]
        mxk = 15
        mdl = SplinesDynamicModel( knots=knots, dynamic=True, maxKnots=mxk, minDistance=0.04 )
        mdl.setLimits( lowLimits=[-10.0], highLimits=[+10.0] )
        
        ep = EvidenceProblem( model=mdl, xdata=t, ydata=y )
        distr = ModelDistribution( scale=0.1 )
        
        ns = NestedSampler( problem=ep, distribution=distr )
        
        ns.verbose = 2
        ns.minimumIterations = 2000
        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10
            ns.minimumIterations = 500
        
        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        ## Plot the evolutie of knots and sample weights
        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        ka = numpy.zeros( ( mxk, len( sl ) ), dtype=float )
        plt.figure( 1, figsize=[14,8] )
        for k,s in enumerate( sl ) :
            n = len( s.model.knots )
            ka[:n,k] = s.model.knots
        for j in range( mxk ) :
            plt.plot( ka[j,:], cc[j%6] )
        wgts = sl.getWeightEvolution()
        mw = max( wgts )
        plt.plot( 100 * wgts / mw, 'k-' )
        plt.xlabel( "Iteration number" )
        plt.ylabel( "Knot position cq. sample weight")
        plt.show()

        plt.figure( 2, figsize=[14,8] )
        plt.plot( t, y, 'k.' )
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-', 'y-']
        cp = ['k*', 'b*', 'r*', 'g*', 'c*', 'm*', 'y*']
        for k in range( 0, 2100, 200 ) :
            kc = k % 7
            s = sl[k]
            mdl = s.model
            kn = mdl.knots
            plt.plot( kn, [-1+k/1000]*len( kn ), cp[kc] )
            plt.plot( t, mdl( t ), cc[kc] )
            print( fmt( k ) )
            print( fmt( kn, max=None ) )
            print( fmt( mdl.parameters, max= None ), fmt( s.logL ) )

            m = BasicSplinesModel( knots=kn )
            ftr = Fitter( t, m, fixedScale=0.05 )
            par = ftr.fit( y )
            lz0 = ftr.getLogZ( limits=[-10,+10] )
            print( fmt( par, max=None ), fmt( lz0 ) )

            m = BasicSplinesModel( knots=kn )
            ftr = Fitter( t, m )
            par = ftr.fit( y )
            lz0 = ftr.getLogZ( limits=[-10,+10] )
            print( fmt( par, max=None ), fmt( lz0 ) )

            m = BasicSplinesModel( knots=kn )
            ftr = Fitter( t, m )
            par = ftr.fit( y )
            lz0 = ftr.getLogZ( limits=[-10,+10], noiseLimits=[0.01,1] )
            print( fmt( par, max=None ), fmt( lz0 ) )

        plt.show()

    def test6( self ) :
        npt = 401

        t, y = self.makeData()

        t = numpy.linspace( 0, 100, npt, dtype=float )
        ym = 2 * numpy.sin( 2 * math.pi * numpy.exp( t / 60 ) +1 ) * numpy.exp( -0.02 * t )
        y = numpy.random.seed( 12345 )
        y = ym + numpy.random.randn( npt ) * 0.05

        kn = [0,25,50,60,70,80,90,100]  
        kn = [0.000,17.254,33.456,42.755,61.243,62.538,74.369,89.854,93.388,100.000]

        m = BasicSplinesModel( knots=kn )
        ftr = Fitter( t, m, fixedScale=0.05 )
        par = ftr.fit( y )
        lz0 = ftr.getLogZ( limits=[-10,+10] )
        print( fmt( par, max=None ), fmt( lz0 ) )

        m = BasicSplinesModel( knots=kn )
        ftr = Fitter( t, m )
        par = ftr.fit( y )
        lz0 = ftr.getLogZ( limits=[-10,+10] )
        print( fmt( par, max=None ), fmt( lz0 ) )

        m = BasicSplinesModel( knots=kn )
        ftr = Fitter( t, m )
        par = ftr.fit( y )
        lz0 = ftr.getLogZ( limits=[-10,+10], noiseLimits=[0.01,1] )
        print( fmt( par, max=None ), fmt( lz0 ) )

    def test7( self ) :
        print( "7  Test Dynamic Modifiable Splines: evidence unknown scale" )

        t, y = self.makeData()

        knots =[0, 100]
        mxk = 15
        mdl = SplinesDynamicModel( knots=knots, dynamic=True, maxKnots=mxk, minDistance=0.01 )
        mdl.setLimits( lowLimits=[-10.0], highLimits=[+10.0] )
        
        ep = EvidenceProblem( model=mdl, xdata=t, ydata=y )
        distr = ModelDistribution( limits=[0.01,1] )
        #ns = NestedSampler( t, mdl, y, seed=1235, engines=eng, distribution=distr )
        #print( ns.problem )
        #print( ns.model )
        #print( fmt( ns.xdata ) )
        #print( fmt( ns.ydata ) )
        
        ns = NestedSampler( problem=ep, distribution=distr )
        
        print( ns.problem )
        print( ns.model )
        print( fmt( ns.xdata ) )
        print( fmt( ns.ydata ) )
        ns.verbose = 2

        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10
            ns.minimumIterations = 500 
       
        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        ## Plot the evolutie of knots and sample weights
        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        ka = numpy.zeros( ( mxk, len( sl ) ), dtype=float )
        plt.figure( 1, figsize=[14,8] )
        for k,s in enumerate( sl ) :
            n = len( s.model.knots )
            ka[:n,k] = s.model.knots
        for j in range( mxk ) :
            plt.plot( ka[j,:], cc[j%6] )
        wgts = sl.getWeightEvolution()
        mw = max( wgts )
        plt.plot( 100 * wgts / mw, 'k-' )
        plt.xlabel( "Iteration number" )
        plt.ylabel( "Knot position cq. sample weight")
        plt.show()

        plt.figure( 2, figsize=[14,8] )
        plt.plot( t, y, 'k.' )
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-', 'y-']
        cp = ['k*', 'b*', 'r*', 'g*', 'c*', 'm*', 'y*']
        for k in range( 0, 2100, 200 ) :
            kc = k % 7
            s = sl[k]
            mdl = s.model
            kn = mdl.knots
            plt.plot( kn, [-1+k/1000]*len( kn ), cp[kc] )
            plt.plot( t, mdl( t ), cc[kc] )
            print( fmt( k ) )
            print( fmt( kn, max=None ) )
            print( fmt( mdl.parameters, max= None ), fmt( s.logL ) )

        plt.show()

    def test8( self ) :
        print( "8  Test Dynamic Modifiable Splines: evidence scale = 1.0" )

        t,y = self.makeData()

        knots =[0, 100]
        mxk = 15
        mdl = SplinesDynamicModel( knots=knots, dynamic=True, maxKnots=mxk, minDistance=0.01 )
        mdl.setLimits( lowLimits=[-10.0], highLimits=[+10.0] )
        
        ep = EvidenceProblem( model=mdl, xdata=t, ydata=y )
        distr = ModelDistribution( )
        
        ns = NestedSampler( problem=ep, distribution=distr )
        
        ns.verbose = 2

        ## Comment next if-statement out for a full run of NestedSampler
        if not self.dofull :
            ns.ensemble = 10
            ns.minimumIterations = 500

        
        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        ## Plot the evolutie of knots and sample weights
        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        ka = numpy.zeros( ( mxk, len( sl ) ), dtype=float )
        plt.figure( 1, figsize=[14,8] )
        for k,s in enumerate( sl ) :
            n = len( s.model.knots )
            ka[:n,k] = s.model.knots
        for j in range( mxk ) :
            plt.plot( ka[j,:], cc[j%6] )
        wgts = sl.getWeightEvolution()
        mw = max( wgts )
        plt.plot( 100 * wgts / mw, 'k-' )
        plt.xlabel( "Iteration number" )
        plt.ylabel( "Knot position cq. sample weight")
        plt.show()

    def suite( cls ):
        return unittest.TestCase.suite( TestModifiable.__class__ )


if __name__ == '__main__':
    unittest.main()




import numpy as numpy
import math
from . import Tools
from .Formatter import formatter as fmt

from .OrderEngine import OrderEngine

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Alpha"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester

class NearEngine( OrderEngine ):
    """
    The NearEngine searches the nearest neighbour (NN) for a point and
    goes there first.

    This is NOT a random engine as it mostly(??) steps uphill.

    It belongs to the class of generalized travelling salesman problems
    where the parameters of the problem is an ordered list.

    The walker is kept when the logLikelihood > lowLhood

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, seed=4213, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : SampleList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : NearEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, copy=copy, seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return NearEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "NearEngine" )

    #  *********EXECUTE***************************************************
    def executeOnce( self, kw, lowLhood, dims=[0,1] ) :
        """
        Execute the NearEngine one time.

        TBC: This engine seems to take a lot of CPU.

        Parameters
        ----------
        kw : int
            id of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        dims : list of 2 ints
            dimensions to process over
        """
        walker = self.walkers[kw]

        problem = walker.problem
        np = problem.npars
        param = walker.allpars

        ks = self.rng.randint( np )
        ptry = numpy.roll( param, ks )


        xd = problem.xdata
        ksel = [0,1]
        ksel = ptry[ksel]
        dmin = problem.distance( xd[ksel,:], [0,1] )[0]         # need only one distance
#        print( 0, fmt( ksel ), fmt( dmin ), fmt( dmin ) )
        
        kmin = 1
        for k in range( 2, np ) :
            ksel[1] = ptry[k]
            d = problem.distance( xd[ksel,:], [0,1] )[0]
#            print( k, fmt( ksel ), fmt( d ), fmt( dmin ) )
            if d < dmin :
                dmin = d
                kmin = k

        if kmin == 1 or kmin == np - 1 :        ## it is already the NN
            self.reportFailed()
            return 0


        ## make the NN the last one
        pkm = ptry[kmin]
        ptry[kmin:-1] = ptry[kmin+1:]
        ptry[-1] = pkm

        ptry = numpy.roll( ptry, -ks )

        Ltry = self.errdis.logLikelihood( problem, ptry )

        if self.verbose > 4 :
            print( "Near     ", ks, kmin, fmt( dmin ), fmt( lowLhood ), fmt( Ltry ) )
            print( fmt( param, max=None, format='%3d' ) )
            print( fmt( ptry, max=None, format='%3d' ) )

        if Ltry >= lowLhood:
            self.reportSuccess( )
            self.setWalker( walker.id, problem, ptry, Ltry )
            t = 1

        else :
            self.reportReject()
            t = 0

        return t                        # nr of succesfull steps





# run with : python3 -m unittest TestDistanceCostFunction

import unittest
import os
import numpy as numpy
from astropy import units
from numpy.testing import assert_array_equal as assertAE
from numpy.testing import assert_array_almost_equal as assertAAE
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma


__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for solving the SalesmanProblem with NestedSampler.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def initProblem( self, np=10, random=False ):

        n2 = np * np
        if random :
            numpy.random.seed( 12345 )
            x0 = numpy.random.rand( n2 ) * np
            x1 = numpy.random.rand( n2 ) * np
        else :
            x0 = numpy.arange( n2, dtype=int ) % np
            x1 = numpy.arange( n2, dtype=int ) // np
        xdata = numpy.append( x0, x1 ).reshape( 2, n2 ).transpose()
        xdata = numpy.array( xdata, dtype=float )

        problem = SalesmanProblem( xdata )

#        numpy.set_printoptions( precision=3, suppress=True )
        return problem

    def test1( self ):
        print( "\n   DistanceCostFunction\n" )

        problem = self.initProblem( random=True )

        pars = numpy.random.permutation( problem.npars )

        print( fmt( problem.xdata, tail=3 ) )
        print( fmt( pars, tail=3 ) )

        swt = problem.sumweight
        dist = problem.result( pars )
        sd1  = numpy.sum( dist )

        print( fmt( dist, tail=3 ) )
        print( fmt( swt ), fmt( sd1 ) )
        self.assertTrue( swt == 100 )
#        assertAAE( sd1, 502.303, 3 )

        dcf = DistanceCostFunction()
        lh = dcf.logLikelihood( problem, pars )
        la = dcf.logLikelihood_alt( problem, pars )
        print( "logL  ", fmt( lh ), fmt(la ) )

        problem.xdata *= 100
        swt = problem.sumweight
        dist = problem.result( pars )
        sd2  = numpy.sum( dist )
        dcf = DistanceCostFunction()

        print( fmt( swt ), fmt( sd2 ) )
        lh = dcf.logLikelihood( problem, pars )
        la = dcf.logLikelihood_alt( problem, pars )
        print( "logL  ", fmt( lh ), fmt(la ) )

        self.assertTrue( swt == 100 )
        assertAAE( sd2, ( 100 * sd1 ) )

        problem.xdata /= 100
        problem.weights = numpy.array( [2.0] * 100 )

        swt = problem.sumweight
        dist = problem.result( pars )
        sd3  = numpy.sum( dist )
        dcc = dcf.copy()

        print( fmt( swt ), fmt( sd3 ) )
        lh = dcc.logLikelihood( problem, pars )
        la = dcc.logLikelihood_alt( problem, pars )
        print( "logL  ", fmt( lh ), fmt(la ) )

        self.assertTrue( dcc.acceptWeight() )
        self.assertTrue( swt == 200 )
        assertAAE( sd3, ( 2 * sd1 ) )

        problem.weights = numpy.array( [1.0,2.0] * 50 )

        swt = problem.sumweight
        dist = problem.result( pars )
        sd3  = numpy.sum( dist )
        dcc = dcf.copy()

        print( fmt( swt ), fmt( sd3 ) )
        lh = dcc.logLikelihood( problem, pars )
        la = dcc.logLikelihood_alt( problem, pars )
        print( "logL  ", fmt( lh ), fmt(la ) )

        self.assertTrue( dcc.acceptWeight() )
        self.assertTrue( swt == 150 )
#        assertAAE( sd3, ( 1.5 * sd1 ) )


    def suite( cls ):
        return unittest.TestCase.suite( TestDistanceCostFunction.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from .BaseFitter import BaseFitter

from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2020 Do Kester

class Fitter( BaseFitter ):
    """
    Fitter for linear models.

    The Fitter class is to be used in conjunction with Model classes.

    The Fitter class and its descendants fit data to a model. Fitter itself
    is the variant for linear models, ie. models linear in its parameters.

    Examples
    --------
    # assume x and y are numpy.asarray data arrays:
    >>> x = numpy.arange( 100 )
    >>> y = numpy.arange( 100 ) // 4        # digitization noise
    >>> poly = PolynomialModel( 1 )         # line
    >>> fitter = Fitter( x, poly )
    >>> param = fitter.fit( y )
    >>> stdev = fitter.stdevs               # stdevs on the parameters
    >>> chisq = fitter.chisq
    >>> scale = fitter.scale                # noise scale
    >>> yfit  = fitter.getResult( )         # fitted values
    >>> yfit  = poly( x )                   # same as previous
    >>> yband = fitter.monteCarloError( )        # 1 sigma confidence region


    Limitations
    -----------
    1. The Fitter does not work with limits.
    2. The calculation of the evidence is an Gaussian approximation which is
       only exact for linear models with a fixed scale.

    Author  Do Kester

    """

    def __init__( self, xdata, model, map=False, keep=None, fixedScale=None ):
        """
        Create a new Fitter, providing xdatas and model.

        A Fitter class is defined by its model and the input vector (the
        independent variable). When a fit to another model and/or another
        input vector is needed a new object should be created.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
        map : bool (False)
            When true, the xdata should be interpreted as a map.
            The fitting is done on the pixel indices of the map,
            using ImageAssistant
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values of keep will be used by the Fitter as long as the Fitter exists.
            See also `fit( ..., keep=dict )`
        fixedScale : float
            the fixed noise scale

        """
        super( Fitter, self ).__init__( xdata, model, map=map, keep=keep, fixedScale=fixedScale )

    def fit( self, ydata, weights=None, keep=None, plot=False ):
        """
        Return model parameters fitted to the data, including weights.

        For Linear models the matrix equation

            H * p = &beta;

        is solved for p. H is the Hessian matrix ( D * w * D^T )
        and &beta; is the inproduct of the data with the D, design matrix.

            &beta; = y * w * D^T

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data ( = 1.0 / sigma^2 )
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
            The values will override those at initialization.
            They are only used in this call of fit.
        plot : bool
            Plot the results

        Raises
        ------
            ValueError when ydata or weights contain a NaN

        """
        fitIndex, ydata, weights = self.fitprolog( ydata, weights=weights, keep=keep )

        if self.model.isNullModel() :
            self.chiSquared( ydata, weights )
            return numpy.asarray( 0 )

        hessian = self.getHessian( weights=weights, index=fitIndex )
        ydatacopy = ydata.copy( )
        # subtract influence of fixed parameters on the data
        if fitIndex is not None :
            fxpar = numpy.copy( self.model.parameters )
            fxpar[fitIndex] = 0.0
            ydatacopy = numpy.subtract( ydatacopy, self.model.result( self.xdata, fxpar) )

        if weights is not None :
            ydatacopy *= weights
        if hasattr( self, "normdfdp" ) :
            ydatacopy = numpy.append( ydatacopy, self.normdata * self.normweight )

        vector = self.getVector( ydatacopy, index=fitIndex )
#        print( fmt( hessian ) )
        params = numpy.linalg.solve( hessian, vector )

        params = self.insertParameters( params, index=fitIndex )
        self.model.parameters = params
        self.chiSquared( ydata, params=params, weights=weights)

        self.fitpostscript( ydata, plot=plot )

        return params

    def __str__( self ):
        """ Return the name of the fitter. """
        return "Fitter"





import numpy as numpy
import math

from .ScaledErrorDistribution import ScaledErrorDistribution
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester


class GaussErrorDistribution( ScaledErrorDistribution ):
    """
    To calculate a Gauss likelihood.

    For one residual, x, it holds

        L( x ) = 1 / &sqrt;( 2 &pi; s^2 ) exp( - 0.5 ( x / s )^2 )

    where s is the scale.
    s is a hyperparameter, which might be estimated from the data.

    The scale s is also the square root of the variance of this error distribution.

    The function is mostly used to calculate the likelihood L over N residuals,
    or easier to use log likelihood, logL.

        logL = log( N / ( sqrt( 2 &pi; ) s )  ) - 0.5 &sum;( x / s ) ^ 2

    Using weights this becomes:

        logL = log( &sum;( w ) / ( sqrt( 2 &pi; ) s )  ) - 0.5 &sum;( w ( x / s ) ^ 2 )


    Author       Do Kester.

    """
    LOG2PI = math.log( 2 * math.pi )


    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, limits=None, copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is *not* fixed.

        copy : GaussErrorDistribution
            distribution to be copied.

        """
        super( GaussErrorDistribution, self ).__init__(  scale=scale, limits=limits, copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return GaussErrorDistribution( copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always true for this distribution.
        """
        return True

    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        return self.getGaussianScale( problem, allpars=allpars )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ) :
        """
        Return the log( likelihood ) for a Gaussian distribution.

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem

        """
        self.ncalls += 1

        scale = allpars[-1]
        chisq = self.getChisq( problem, allpars ) / ( scale * scale )
        return ( - problem.sumweight * ( 0.5 * self.LOG2PI + math.log( scale ) ) -
                       0.5 * chisq )

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        scale = allpars[-1]

        res2 = -0.5 * problem.weightedResSq( allpars[:-1], mockdata=mockdata )

        res2 /= ( scale * scale )
        if problem.weights is None :
            res2 -= ( 0.5 * self.LOG2PI + math.log( scale ) )
        else :
            res2 -= ( 0.5 * self.LOG2PI + math.log( scale ) ) * problem.weights
        return res2


    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters in fitIndex.

        Alternate calculation

        Parameters
        ----------
        problem : Problem
            to be solved.
        allpars : array_like
            (hyper)parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """

        self.nparts += 1                    ## counts calls to partialLogL

        scale = allpars[-1]
        s2 = scale * scale
        ( res2, res ) = problem.weightedResSq( allpars[:-1], extra=True )

        dM = problem.partial( allpars[:-1] )

        dL = numpy.zeros( len( fitIndex ), dtype=float )
        i = 0
        for  k in fitIndex :
            if k >= 0 :
                dL[i] = numpy.sum( res * dM[:,k] ) / s2
                i += 1
            else :
                dL[-1] = ( numpy.sum( res2 ) / s2 - problem.sumweight ) / scale

        return dL


    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ) :
        """
        Return the partial derivative all elements of the log( likelihood )
        to the parameters in fitIndex.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            (hyper)parameters of the problem
        fitIndex : array_like of int
            indices of allpars to fit
        mockdata : array_like
            as calculated for the problem
        """

        np = problem.npars
        param = allpars[:np]

        ( res2, res ) = problem.weightedResSq( allpars[:-1], mockdata=mockdata, extra=True )

        dM = problem.partial( param )

##      TBD import mockdata into partial
#        dM = problem.partial( param, mockdata=mockdata )

        scale = allpars[-1]
        s2 = scale * scale
        for  k in fitIndex:
            if k >= 0 :                         ## the parameters
                yield ( res * dM[:,k] ) / s2
            else :                              ## the scale
                wgt = 1.0 if problem.weights is None else problem.weights
                yield ( res2 / s2 - wgt ) / scale

        return

    def hessianLogL( self, problem, allpars, fitIndex ) :
        """
        Return the hessian of log( likelihood ) to the parameters in fitIndex.

        The hessian is a matrix containing the second derivatives to each
        of the parameters.

             hessian = d^2 logL / dp_i dp_k

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            (hyper)parameters of the problem
        fitIndex : array_like of int
            indices of allpars to fit
        """
        self.nparts += 1                    ## counts calls to partialLogL

        nh = len( fitIndex )
        np = problem.npars
        param = allpars[:np]
        scale = allpars[np]
        s2 = scale * scale

        fi = fitIndex if fitIndex[-1] != np else fitIndex[-1:]

        nf = len( fi )
        design = problem.partial( param )[:,fi]
        design = design.transpose()
        deswgt = design
        if problem.hasWeights() :
            deswgt *= problem.weights

        hessian = numpy.zeros( ( nh, nh ), dtype=float )
        hessian[:nf,:nf] = numpy.inner( design, deswgt ) / s2

        if fitIndex[-1] == np :
            hessian[nf,nf] = 2 * ( problem.ndata - nf ) / s2

        return hessian

    def __str__( self ) :
        return "Gauss error distribution"



# run with : python3 -m unittest TestPrior2
# or :       python3 -m unittest TestPrior2.Test.test1

import unittest
import numpy as numpy
import math

from BayesicFitting import Prior, UniformPrior, JeffreysPrior, ExponentialPrior
from BayesicFitting import LaplacePrior, CauchyPrior, GaussPrior
from BayesicFitting import CircularUniformPrior
from BayesicFitting import PolynomialModel, NestedSampler

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class Test( unittest.TestCase ) :
    """
    Test for Priors
    """

    def test1( self ) :

        print( "===========Gauss Prior Limitation ===========" )

        NP = 100
        x = numpy.zeros( NP, dtype=float )
        y = numpy.random.randn( NP )

        print( numpy.mean( numpy.append( y, [5.0] ) ) )

        mdl = PolynomialModel( 0 )

        mdl.setPrior( 0, GaussPrior( center=4.0, scale=1.0 ) )

        ns = NestedSampler( x, mdl, y, verbose=2 )

        logE = ns.sample()

        print( ns.parameters, logE )

        lim = math.exp( 4 )
        mdl.setPrior( 0, UniformPrior( limits=[-lim,lim] ) )

        ns = NestedSampler( x, mdl, y, verbose=2 )

        logE = ns.sample()

        print( ns.parameters, logE )



    @classmethod
    def suite( cls ):
        return ConfiguredTestCase.suite( PriorTest.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .NonLinearModel import NonLinearModel
from .Kepplers2ndLaw import Kepplers2ndLaw
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2018 - 2022 Do Kester

class StellarOrbitModel( NonLinearModel ):
    """
    Model for the radial velocity variations of a star caused by a orbiting planet.
.
    The algorithm was taken from
        Cory Boule etal. (2017) J. of Double Star Observations Vol 13 p.189.
        http://www.jdso.org/volume13/number2/Harfenist_189-199.pdf

    p_0 : e     eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
    p_1 : a     semi major axis (>0)
    p_2 : P     period of the orbit (>0)
    p_3 : T     phase since periastron passage (0<p_3<2pi)
    p_4 : i     inclination of the orbit wrt sky (0<i<pi; 0 = pi = in sky plane)
    p_5 : Omega position angle from North to the line of nodes (0<Omega<pi; 0 = north )
    p_6 : omega longitude from the node (in p_5) to the periastron (0<omega<2pi; 0 = periastron in node )

    Due to the fact that the orbit can be mirrored in the sky plane, one of p_5 or p_6
    has to be limited to [0,pi] and the other to [0,2pi].

    The parameters are initialized at [0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0].
    It is a non-linear model.

    This class uses @Kepplers2ndLaw to find the radius and anomaly.

    Attributes
    ----------
    keppler : Kepplers2ndLaw()
        to calculate the radius and true anomaly
    ndout : int
        The number of outputs is 2. Use @MultipleOutputProblem.
    spherical : bool
        if True return the results in spherical coordinates.
    cyclic : { 1 : 2*pi }
        Only if spherical, indicating that result[:,1] is cyclic.

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Examples
    --------
    >>> sm = StellarOrbitModel( )
    >>> print( sm.npars )
    7


    """
    def __init__( self, copy=None, spherical=True, **kwargs ):
        """
        Radial velocity model.

        Number of parameters is 5:

        Parameters
        ----------
        copy : StellarOrbitModel
            model to copy
        spherical : bool (True)
            produce output in sperical coordinates (rho,phi)
            otherwise in rectilinear coordinates (x,y)
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.

        """
        param = [0.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]
        names = ["eccentricity", "semimajoraxis", "period", "phase since periastron",
                  "inclination", "north2nodes", "nodes2periastron"]

        setatt( self, "ndout", 2 )
        setatt( self, "spherical", spherical )

        super( StellarOrbitModel, self ).__init__( 7, copy=copy, params=param,
                        names=names, **kwargs )

        setatt( self, "keppler", Kepplers2ndLaw() )

        if spherical :
            setatt( self, "cyclic", {1:2*math.pi} )

    def copy( self ):
        """ Copy method.  """
        return StellarOrbitModel( spherical=self.spherical, copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        p_0 : e     eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
        p_1 : a     semi major axis (>0)
        p_2 : P     period of the orbit (>0)
        p_3 : p     phase since periastron passage (0<p<2pi)
        p_4 : i     inclination of the orbit wrt sky (0<i<pi; 0 = orbit in sky plane)
        p_5 : Omega position angle of the line of nodes
        p_6 : omega longitude of periastron from lines of nodes

        """
        inclin = params[4]
        ascpos = params[5]
        asclon = params[6]

        ( r, v ) = self.keppler.radiusAndTrueAnomaly( xdata, params[:4] )

        vp = v + asclon
        svp = numpy.sin( vp ) * math.cos( inclin )
        cvp = numpy.cos( vp )

        ## theta = angle on sky of stars wrt each other
        theta = numpy.arctan2( svp, cvp ) + ascpos
        ## rho = distance between stars
        rho = r * numpy.sqrt( svp * svp + cvp * cvp )

        result = numpy.zeros( ( len( xdata ), 2 ), dtype=float )
        result[:,0] = rho       ## distance
        result[:,1] = theta     ## positional angle from north counterclock

        if not self.spherical :
            result = self.toRect( result )

        return result

    def toRect( self, rp ):
        """
        Return (x,y) coordinates from (rho,phi)

        Parameters
        ----------
        rp : array
            rp[:,0] : separation of the stars
            rp[:,1] : angle from north (down) CCW to east (right)
        """
        xy = numpy.empty_like( rp )
        xy[:,0] = rp[:,0] * numpy.sin( rp[:,1] )
        xy[:,1] =-rp[:,0] * numpy.cos( rp[:,1] )
        return xy


    def toSpher( self, xy ) :
        """
        Return (rho,phi) coordinates from (x,y)

        Parameters
        ----------
        xy : array
            xy[:,0] : x position
            xy[:,1] : y position
        """
        rp = numpy.empty_like( xy )
        rp[:,0] = numpy.hypot( xy[:,0], xy[:,1] )
        rp[:,1] = numpy.arctan2( xy[:,0], -xy[:,1] )
        return rp

    def XXXderivative( self, xdata, params ):
        return self.baseDerivative( xdata, params )

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative (df/dx) of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        p_0 : e     eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
        p_1 : a     semi major axis (>0)
        p_2 : P     period of the orbit (>0)
        p_3 : p     phase since periastron passage (0<p<2pi)
        p_4 : i     inclination of the orbit wrt sky (0<i<pi; 0 = orbit in sky plane)
        p_5 : O(mega) position angle of the line of nodes (0<Omega<pi)
        p_6 : o(mega) longitude of periastron (0<omega<2pi)

        """
        inclin = params[4]
        ascpos = params[5]
        asclon = params[6]

        dfdx = numpy.zeros( ( len( xdata ), 2 ), dtype=float )

        p = params[:4]
        r, v = self.keppler.radiusAndTrueAnomaly( xdata, p )

        cosE = self.keppler.cosE
        cosE = numpy.where( cosE == -1, -0.99999, cosE )
        sinE = self.keppler.sinE
        drdx, dvdx = self.keppler.drvdx( xdata, p, cosE, sinE )

        vp = v + asclon
        svp = numpy.sin( vp )
        cvp = numpy.cos( vp )
        ci = math.cos( inclin )
        svp2 = svp * svp
        cvp2 = cvp * cvp
        ci2 = ci * ci

        ## theta = angle on sky of stars wrt each other
        theta = numpy.arctan2( svp * ci, cvp ) + ascpos
        dtdv = ci * ( svp2 + cvp2 ) / ( ci2 * svp2 + cvp2 )
        dtdx = dtdv * dvdx

        ## rho = distance between stars
        ssc = numpy.sqrt( svp2 * ci2 + cvp2 )
        rho = r * ssc
        dRdr = ssc
        dRdv = r / ssc * ( svp * cvp * ( ci2 - 1 ) )
        dRdx = dRdr * drdx + dRdv * dvdx

        if self.spherical :
            dfdx[:,0] = dRdx
            dfdx[:,1] = dtdx
        else :
            ct = numpy.cos( theta )
            st = numpy.sin( theta )
            dfdx[:,0] = +st * dRdx + rho * ct * dtdx
            dfdx[:,1] = -ct * dRdx + rho * st * dtdx

        return dfdx


    def XXXpartial( self, xdata, params, parlist=None ) :
        return self.basePartial( xdata, params, parlist=parlist )

    def basePartial( self, xdata, params, parlist=None ) :
        """
        Returns the partials at the xdata value.
        <br>
        The partials are x ( xdata ) to degree-th power.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters. (not used for linear models)
        parlist : array_like
            list of indices active parameters (or None for all)

        p_0 : e     eccentricity of the elliptic orbit (0<e<1; 0 = circular orbit)
        p_1 : a     semi major axis
        p_2 : P     period of the orbit (>0)
        p_3 : p     phase since periastron passage (0<p<2pi)
        p_4 : i     inclination of the orbit wrt sky (0<i<pi; 0 = orbit in sky plane)
        p_5 : O     position angle of the line of nodes
        p_6 : o     longitude of periastron

        """
        inclin = params[4]
        ascpos = params[5]
        asclon = params[6]

        p = params[:4]
        r, v = self.keppler.radiusAndTrueAnomaly( xdata, p )

        cosE = self.keppler.cosE
        cosE = numpy.where( cosE == -1, -0.99999, cosE )
        sinE = self.keppler.sinE
        E = self.keppler.eccAnomaly
        drde, drda, drdP, drdp, dvde, dvdP, dvdp = self.keppler.drvdpar( xdata, p,
                                                   E, cosE, sinE )

        vp = v + asclon
        svp = numpy.sin( vp )
        cvp = numpy.cos( vp )
        ci = math.cos( inclin )
        si = math.sin( inclin )
        svp2 = svp * svp
        cvp2 = cvp * cvp
        ci2 = ci * ci

        partial0 = numpy.zeros( ( Tools.length( xdata ), self.npbase ), dtype=float )
        partial1 = numpy.zeros( ( Tools.length( xdata ), self.npbase ), dtype=float )

        ## theta = angle on sky of stars wrt each other
        theta = numpy.arctan2( svp * ci, cvp ) + ascpos
        dTdv = ci * ( svp2 + cvp2 ) / ( ci2 * svp2 + cvp2 )

        ## rho = distance between stars
        ssc = numpy.sqrt( svp2 * ci2 + cvp2 )
        rho = r * ssc
        dRdr = ssc
        dRdv = r / ssc * ( ci2 - 1 ) * svp * cvp


        part0 = { 0 : ( lambda: dRdr * drde + dRdv * dvde ),
                  1 : ( lambda: dRdr * drda ),
                  2 : ( lambda: dRdr * drdP + dRdv * dvdP ),
                  3 : ( lambda: dRdr * drdp + dRdv * dvdp ),
                  4 : ( lambda: -r / ssc * svp2 * ci * si ),
                  5 : ( lambda: 0 ),
                  6 : ( lambda: dRdv ) }

        part1 = { 0 : ( lambda: dTdv * dvde ),
                  1 : ( lambda: 0 ),
                  2 : ( lambda: dTdv * dvdP ),
                  3 : ( lambda: dTdv * dvdp ),
                  4 : ( lambda: -cvp * svp * si / ( cvp2 + svp2 * ci2 ) ),
                  5 : ( lambda: 1 ),
                  6 : ( lambda: dTdv ) }

        if not self.spherical :
            dRde = part0[0]()
            dRda = part0[1]()
            dRdP = part0[2]()
            dRdp = part0[3]()
            dRdi = part0[4]()
            dRdO = 0
            dRdo = part0[6]()

            dTde = part1[0]()
            dTda = 0
            dTdP = part1[2]()
            dTdp = part1[3]()
            dTdi = part1[4]()
            dTdO = 1
            dTdo = part1[6]()

            ct = numpy.cos( theta )
            st = numpy.sin( theta )
            rst = rho * st
            rct = rho * ct
            ct *= -1

            part0 = { 0 : ( lambda: st * dRde + rct * dTde ),
                      1 : ( lambda: st * dRda + rct * dTda ),
                      2 : ( lambda: st * dRdP + rct * dTdP ),
                      3 : ( lambda: st * dRdp + rct * dTdp ),
                      4 : ( lambda: st * dRdi + rct * dTdi ),
                      5 : ( lambda: st * dRdO + rct * dTdO ),
                      6 : ( lambda: st * dRdo + rct * dTdo ) }
            part1 = { 0 : ( lambda: ct * dRde + rst * dTde ),
                      1 : ( lambda: ct * dRda + rst * dTda ),
                      2 : ( lambda: ct * dRdP + rst * dTdP ),
                      3 : ( lambda: ct * dRdp + rst * dTdp ),
                      4 : ( lambda: ct * dRdi + rst * dTdi ),
                      5 : ( lambda: ct * dRdO + rst * dTdO ),
                      6 : ( lambda: ct * dRdo + rst * dTdo ) }


        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial0[:,k] = part0[kp]()
            partial1[:,k] = part1[kp]()


        return ( partial0, partial1 )


    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        return str( "StellarOrbit " )

    def baseParameterUnit( self, k ):
        """
        Return the unit of a parameter. (TBC)

        Parameters
        ----------
        k : int
            the kth parameter.

        """

        if k == 0:
            return units.Unit( )
        elif k == 1 :
            return self.yUnit
        elif k == 2 :
            return 1.0 / self.xUnit
        return units.Unit( units.si.rad )





import numpy as numpy

from .Dynamic import Dynamic
from .Engine import Engine
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester

class BirthEngine( Engine ):
    """
    The BirthEngine adds a new component to the model.

    Only for Models that are Dynamic.
    The birth rate is governed by the growth-prior in the Dynamic.

    The member is kept when the logLikelihood > lowLhood.

    Author       Do Kester.

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, slow=None, copy=None, seed=23455, verbose=0 ) :
        """
        Constructor.

        Parameters
        ----------
        walkers : list of Walker
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        slow : None or int > 0
            Run this engine every slow-th iteration. None for all.
        copy : BirthEngine
            to be copied
        seed : int
            for random number generator

        """
        super( ).__init__( walkers, errdis, slow=slow, copy=copy,
                    seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return BirthEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "BirthEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by adding a component and diffusing the parameters.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()            ## work on local copy.

        nhyp = self.errdis.nphypar
        walker.problem.model.parameters = walker.allpars[:-nhyp] if nhyp > 0 else walker.allpars

        problem = walker.problem
        allp = walker.allpars
        ftry = walker.fitIndex

        if self.verbose > 4 :
            print( "BEN0  ", walker.id, walker.parent, fmt( allp, max=None ), len( ftry ) )

        off = 0
        model = problem.model
        model.parameters = allp[:model.npars]

        while model is not None and not isinstance( model, Dynamic ) :
            off += model.npbase
            model = model._next

        nc = model.ncomp
        np = model.npbase

        if self.verbose > 4 :
            print( "BEN1  ", walker.id, nc, np, len( allp ), len( ftry ) )

        if not ( nc < model.growPrior.unit2Domain( self.rng.rand() ) and
                 model.grow( offset=off, rng=self.rng ) ):
            self.reportReject()
            return 0

        dnp = model.npbase - np         # parameter change
        ftry = model.alterFitindex( ftry, np, dnp, off )
        ptry = problem.model.parameters          ## list of grown parameters
        if nhyp > 0 :
            ptry = numpy.append( ptry, allp[-nhyp:] )

        loc = model.location            # where the extra parameters are inserted
        del( model.location )           # not needed anymore

        t = 0
        k0 = off + loc
        k1 = k0 + dnp
        klst = [k for k in range( k0, k1 )]

        while t < self.maxtrials :

            Ltry = self.errdis.logLikelihood( problem, ptry )

            if Ltry >= lowLhood:
                self.reportSuccess()
                update = len( self.walkers ) if append else kw
                self.setWalker( update, problem, ptry, Ltry, fitIndex=ftry )
                wlkr = self.walkers[update]
                wlkr.check( nhyp=nhyp )
                return dnp
            else :
                uval = self.rng.rand( dnp )
                ptry[klst] = self.unit2Domain( problem, uval, kpar=klst )

                self.reportReject()
                t += 1

#        print( "Birth   ", walker.id, fmt( ptry, max=None ) )

        self.reportFailed()

        return 0





import numpy as numpy
import math

from .HyperParameter import HyperParameter
from . import Tools
from .Tools import setAttribute as setatt

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.0.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2022 Do Kester


class ErrorDistribution( object ):
    """
    ErrorDistribution defines general methods for a error distribution.

    Error distributions are used to calculate the likelihoods.

    Author       Do Kester.

    Attributes
    ----------
    hyperpar : HyperParameter
        hyperparameter for the error distribution
    deltaP : float
        delta for calculating numerical derivatives
    ncalls : int
        number of calls to the logLikelihood
    nparts : int
        number of calls to the partial of the logLikelihood
    sumweight : float
        sum over the weights or ndata
    ndata : int
        number of points in data
    hypar : [float]
        list of values for the hyperparameters
    nphypar : int
        number of hyper parameters in this error distribution
    constrain : None or callable
        None:     Use logLikelihood as is
        callable: logL = func( logL, problem, allpars, lowLhood )
                  returning a (modified) value of the logLikelihood.
    """

    PARNAMES = ["hypar"]

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, fixed=None, constrain=None, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
        constrain : None or callable
            function as: func( logL, problem, allpars )
            returning a (modified) value of the logLikelihood.
        copy : ErrorDistribution
            distribution to be copied.

        Raise
        -----
        ValueError when constrain is not a callable method.

        """
        super( ErrorDistribution, self ).__init__()
        self.ncalls = 0
        self.nparts = 0
        if copy is None :
            self.hyperpar = []
            self.deltaP = 0.000001
            self.fixed = self.keepFixed( fixed )
            self.constrain = constrain
        else :                          ### TBC do we need copies here ????
            self.hyperpar = copy.hyperpar
            self.deltaP = copy.deltaP
            self.fixed = self.keepFixed( copy.fixed )
            self.constrain = copy.constrain

    def copy( self ):
        """ Return copy of this.  """
        return ErrorDistribution( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        if name == "constrain" :
            if hasattr( self, "logLdata" ) :
                ## Choose one of 2 likelihood functions in "logLikelihood"
                if value is None :
                    setatt( self, "logLikelihood", self.logLhood )
                elif callable( value ) :
                    setatt( self, "logLikelihood", self.logCLhood )
                else :
                    raise ValueError( "Constrain is not a callable function" )

            setatt( self, "constrain", value )
            return

        key0 = ["hyperpar", "fixed"]
        keys = {"hyperpar": HyperParameter}
        key1 = {"deltaP": float, "lowLhood": float, "ncalls": int, "nparts": int, "fixed": dict }
        if ( Tools.setNoneAttributes( self, name, value, key0 ) or
             Tools.setListOfAttributes( self, name, value, keys ) or
             Tools.setSingleAttributes( self, name, value, key1 ) ) :
            pass
        else :
            raise AttributeError(
                "Object has no attribute " + name + " of type " + str( value.__class__ ) )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == 'nphypar' :
            return len( self.hyperpar )
        elif name == "hypar" :
            return numpy.asarray( [s.hypar for s in self.hyperpar], dtype=float )
        else :
            raise AttributeError( "Unknown attribute " + name )

        return None

    def getGaussianScale( self, problem, allpars=None ) :
        """
        Return the noise scale.

        *** Gaussian approximation ***

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        chi = self.getChisq( problem, allpars=allpars )
        return math.sqrt( chi / problem.sumweight )


    def getResiduals( self, problem, allpars=None ):
        """
        Return residuals: ydata - model.result

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        if allpars is None :
            param = problem.model.parameters
        else :
            nh = self.nphypar
            param = allpars[:-nh]

        return problem.residuals( param )


    def getChisq( self, problem, allpars=None ):
        """
        Return chisq

        *** Gaussian approximation ***

        Sum over the (weighted) squared residuals

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        if allpars is None :
            param = problem.model.parameters
        else :
            nh = self.nphypar
            param = allpars[:-nh]

        res2 = problem.weightedResSq( param )

        return numpy.sum( res2 )

    def toSigma( self, scale ) :
        """
        Return sigma, the squareroot of the variance.

        Parameter
        --------
        scale : float
            the scale of this distribution.

        Return default value : scale
        """
        return scale

    def isBound( self ) :
        """
        True when all priors of its (hyper)parameters are bound
        """
        bound = True
        if self.hyperpar is None :
            return bound

        for sp in self.hyperpar :
            bound = bound and sp.prior.isBound()

        return bound

    def acceptWeight( self ):
        """ True if the distribution accepts weights.  """
        return False

    def keepFixed( self, fixed=None ) :
        """
        Keeps (hyper)parameters fixed at the provided values.

        1. Repeated calls start from scratch.<br>
        2. Reset with keepFixed( fixed=None )

        Parameters
        ----------
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.
        """
        if self.hyperpar is None :
            return None

        # Reset all fixed to False
        for s in self.hyperpar :
            s.fixed = False

        if fixed is None :
            return None

        for f in fixed.keys() :
            self.hyperpar[f].fixed = True
            self.hyperpar[f].hypar = fixed[f]

        return fixed


    def setPriors( self, priors ) :
        """
        Set priors on the hyper parameter(s).

        Parameters
        ----------
        priors  : (list of) Prior
            prior distribution for the hyperparameters
        """
        for spp,pr in zip( self.hyperpar, priors ) :
            spp.setPrior( pr )

    def setLimits( self, limits ) :
        """
        Set limits on the hyper parameter(s).

        Parameters
        ----------
        limits : [low,high]
            low : float or array_like
                low limits
            high : float or array_like
                high limits
        """
        if self.nphypar == 1 :
            self.hyperpar[0].setLimits( limits )
            return
        low = limits[0]
        high = limits[1]
        for spp,lo,hi in zip( self.hyperpar, low, high ) :
            spp.setLimits( [lo,hi] )

    def domain2Unit( self, dval, ks ) :
        """
        Return value in [0,1] for the selected parameter.
        Parameters
        ----------
        dval : float
            hyper parameter value in domain
        ks : int
            selecting index
        """
        try :
            return self.hyperpar[ks].domain2Unit( dval )
        except :
            return 0.0

    def unit2Domain( self, uval, ks ) :
        """
        Return domain value for the selected parameter.
        Parameters
        ----------
        uval : float
            unit value of hyper parameter
        ks : int
            selecting index
        """
        try :
            return self.hyperpar[ks].unit2Domain( uval )
        except :
            return 0.0

    #  *********LIKELIHOODS***************************************************

    def logCLhood( self, problem, allpars ):
        """
        Return the constrained log( likelihood ).

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        """
        logL = self.logLhood( problem, allpars )

        return self.constrain( logL, problem, allpars, self.lowLhood )

    def logLhood( self, problem, allpars ):
        """
        Return the log( likelihood ).

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        """
        self.ncalls += 1

        return numpy.sum( self.logLdata( problem, allpars ) )


    def partialLogL( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        self.nparts += 1                ## counts calls tp partialLogL
        mock = problem.result( allpars[:problem.model.npars] )

        try :
            pg = self.nextPartialData( problem, allpars, fitIndex, mockdata=mock )
            np = len( fitIndex )
            dL = numpy.zeros( np, dtype=float )
            for k in range( np ) :
                dL[k] = numpy.sum( next( pg ) )

            try :
                pg.close()
            except :
                pass

            return dL
        except :
#            raise
            print( "Using numeric partialLogL." )
            return self.numPartialLogL( problem, allpars, fitIndex )


    def partialLogL_alt( self, problem, allpars, fitIndex ) :
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Alternative calculation.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        return self.numPartialLogL( problem, allpars, fitIndex )

    def numPartialLogL( self, problem, allpars, fitIndex ) :
        """
        Return d log( likelihood ) / dp, numerically calculated.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            parameters of the problem
        fitIndex : array_like
            indices of parameters to be fitted

        """
        dL = numpy.zeros( len( fitIndex ), dtype=float )
        p = allpars.copy( )
        i = 0
        for k in fitIndex :
            p[k] = allpars[k] - self.deltaP
            lm = self.logLikelihood( problem, p )
            p[k] = allpars[k] + self.deltaP
            lp = self.logLikelihood( problem, p )
            dL[i] = 0.5 * ( lp - lm ) / self.deltaP
            i += 1
            p[k]= allpars[k]
        return dL


    def updateLogL( self, problem, allpars, parval=None ):
        """
        Return a update of the log( likelihood ) given a change in a few parameter.

        This method provides the opportunity to optimize the logL calculation.
        Providing this one, automatically provides the previous one.
        For now it just refers to logLikelihood() itself.

        Parameters
        ----------
        problem : Problem
            to be solved
        param : array_like
            parameters of the model
        parval : dict of {int : float}
            int index of a parameter
            float (old) value of the parameter
        """
        return self.logLikelihood( problem, allpars )

    def setResult( self ):
        pass

    def __str__( self ) :
        """
        Return a string representation of this class.
        """
        return "Error distribution"

    def hyparname( self, k ) :
        """
        Return name of the hyperparameter

        Parameters
        ----------
        k : int
            index of the hyperparameter
        """
        return self.PARNAMES[k]






# run with : python3 -m unittest TestEvidence2

import unittest
import os
import numpy
import math
from numpy.testing import assert_array_almost_equal as assertAAE

import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt

from FitPlot import plotFit
from FitPlot import plotErrdis
from FitPlot import plotErrdis2d


class TestEvidence2( unittest.TestCase  ) :

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ) :
        print( "====test1============================" )
        nn = 100
        x = numpy.zeros( nn, dtype=float )
        ym = 0.2 + 0.5 * x
        nf = 1.0
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-2,2]

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution( )
        problem = ClassicProblem( pm, xdata=x, ydata=y )
        logz1, maxll = plotErrdis( errdis, problem, limits=limits,
                                    max=0, plot=self.doplot )

        print( "logZ  ", fmt( logz1 ) )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        stdv = ns.stdevs
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( stdv ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 2*dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if self.doplot :
            plt.plot( parevo, numpy.exp( llevo ), 'r,' )

            mxl = numpy.exp( numpy.max( llevo ) ) * 1.2
            plt.plot( [pars,pars], [0.0,mxl], 'b-' )
            plt.plot( [par2,par2], [0.0,mxl], 'r-' )
            plt.plot( [par2,par2]+stdv, [0.0,mxl], 'g-' )
            plt.plot( [par2,par2]-stdv, [0.0,mxl], 'g-' )

            plt.show()

    def test2( self ) :
        print( "====test2============================" )
        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.4 + 0.5 * x
        nf = 0.5
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution( )
        problem = ClassicProblem( pm, xdata=x, ydata=y )

        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, max=0,
                    plot=self.doplot )

        if self.doplot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 1.0 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if self.doplot :
            plt.show()

    def test2_1( self ) :
        print( "====test2_1============================" )
        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.2 + 0.5 * x
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = GaussErrorDistribution( )
        problem = ClassicProblem( pm, xdata=x, ydata=y )

        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, max=0,
                    plot=self.doplot )
        if self.doplot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz0 ) < 1.0 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if self.doplot :
            plt.show()

    def test3( self ) :
        print( "====test3======fixed noisescale======================" )
        plot = self.doplot

        nn = 10
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.2
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        pm = PolynomialModel( 1 )
        bf = Fitter( x, pm, fixedScale=nf )

        pars = bf.fit( y )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        if plot :
            plt.figure( "model" )
            plotFit( x, data=y, model=pm, ftr=bf, truth=ym, show=False )

        errdis = GaussErrorDistribution( scale=nf )
        problem = ClassicProblem( pm, xdata=x, ydata=y )

        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, max=0,
                        plot=plot )

        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        dis = GaussErrorDistribution( scale=nf )
        ns = PhantomSampler( x, model, y, distribution=dis, verbose=0 )

        logE = ns.sample()
        par2 = ns.parameters
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

#        print( logz0 - logz1, logz0 - logz2 )
        self.assertTrue( abs( logz2 - logz0 ) < 2 * dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )

            plt.figure( "model" )			# grab again
            yfit = ns.modelfit
            err = samples.monteCarloError( x )
            plt.plot( x, yfit + err, 'b-' )
            plt.plot( x, yfit - err, 'b-' )

            plt.show()

    def test4( self ) :
        print( "====test4===unknown noisescale=========================" )
        plot = self.doplot

        nn = 100
        x = numpy.arange( nn, dtype=float ) / 50
        ym = 0.4 + 0.0 * x
        nf = 0.5
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [0,1]
        nslim  = [0.1,1.0]

        pm = PolynomialModel( 0 )
        bf = Fitter( x, pm )

        pars = bf.fit( y )
        scale = bf.scale
        logz0 = bf.getLogZ( limits=limits, noiseLimits=nslim )
        logl0 = bf.logLikelihood
        print( "pars  ", fmt( pars ), "  scale  ", fmt( scale ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        if plot :
            plt.figure( "model" )
            plotFit( x, data=y, model=pm, ftr=bf, truth=ym, show=False )

        errdis = GaussErrorDistribution( )
        problem = ClassicProblem( pm, xdata=x, ydata=y )

        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, nslim=nslim,
                        plot=plot )
        if plot :
            plt.plot( pars[0], scale, 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        dis = GaussErrorDistribution( )
        dis.setLimits( nslim )
        ns = PhantomSampler( x, model, y, distribution=dis, verbose=0 )
        ns.step = 1

        logE = ns.sample()
        par2 = ns.parameters
        scl2 = ns.scale
        logz2 = ns.logZ
        dlz2 = ns.logZprecision
        print( "pars  ", fmt( par2 ), "  scale  ", fmt( scl2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < 2 * dlz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        scevo =samples.getScaleEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], scevo, 'k,' )

            plt.figure( "model" )			# grab again
            yfit = ns.yfit
            err = samples.monteCarloError( x )
            plt.plot( x, yfit + err, 'b-' )
            plt.plot( x, yfit - err, 'b-' )

            plt.show()

    def test5_1( self ) :
        self.dotest5( step=1 )
    def test5_2( self ) :
        self.dotest5( step=2 )
    def test5_4( self ) :
        self.dotest5( step=4 )
    def test5_5( self ) :
        self.dotest5( step=5 )
    def test5_10( self ) :
        self.dotest5( step=10 )


    def dotest5( self, step=10 ) :
        print( "====test5====step=%d========================" % step )
        plot = self.doplot

        nn = 10
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.1
        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        y = ym + nf * noise
        limits = [-1,2]

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        s = 0.0
        s2 = 0.0
        mr = 10
        for k in range( mr ) :
            dis = GaussErrorDistribution( scale=0.5 )
            ns = PhantomSampler( x, model, y, distribution=dis, verbose=0, seed=k+12*step )
            ns.step = step

            logE = ns.sample()
            par2 = ns.parameters
            logz2 = ns.logZ
            dlz2 = ns.logZprecision
            s += logz2
            s2 += logz2 * logz2
            print( "pars  ", fmt( par2 ), "  logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        logz = s / mr
        dlz = math.sqrt( s2 / mr - logz * logz )

        print( "Average  ", fmt( logz ), " +- ", fmt( dlz ) )

    def test6( self ) :
        print( "====test6  Laplace ================" )
        plot = self.doplot

        nn = 20
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.5 * x
        nf = 0.9
        numpy.random.seed( 2345 )
        noise = numpy.random.laplace( size=nn )

        y = ym + nf * noise
        limits = [-1,2]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = AmoebaFitter( x, model, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = LaplaceErrorDistribution( scale=nf )
        problem = ClassicProblem( model, xdata=x, ydata=y )
        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, max=0,
                plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, distribution='laplace', seed=8907,
                    verbose=0, rate=0.5 )

        logE = ns.sample()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < 4 * dlz2 )
#        print( logz0 - logz1, logz0 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )

            plt.show()

    def test6_0( self ) :
        print( "====test6_0  Laplace ================" )
        plot = self.doplot

        nn = 20
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 0.0 * x
        nf = 0.9
        numpy.random.seed( 2345 )
        noise = numpy.random.laplace( size=nn )

        y = ym + nf * noise
        limits = [-1,2]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = PowellFitter( x, model, errdis="laplace" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", pars )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        errdis = LaplaceErrorDistribution( scale=nf )
        problem = ClassicProblem( model, xdata=x, ydata=y )
        logz1, logl1 = plotErrdis( errdis, problem, limits=limits, max=0,
                plot=plot )
        if plot :
            plt.plot( pars[0], logl1, 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )


        model = PolynomialModel( 0 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, distribution='laplace', verbose=0 )

        logE = ns.sample()

        par2 = ns.parameters
        stdv = ns.stdevs
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( stdv ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

#        self.assertTrue( abs( logz2 - logz0 ) < dlz2 )
        print( logz1 - logz2 )

        samples = ns.samples
        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo, numpy.exp( llevo ), 'r,' )

            mxl = numpy.exp( numpy.max( llevo ) ) * 1.2
            plt.plot( [pars,pars], [0.0,mxl], 'b-' )
            plt.plot( [par2,par2], [0.0,mxl], 'r-' )
            plt.plot( [par2,par2]+stdv, [0.0,mxl], 'g-' )
            plt.plot( [par2,par2]-stdv, [0.0,mxl], 'g-' )

            plt.show()


    def test7( self ) :
        print( "====test7  Poisson ================" )
        plot = self.doplot

        nn = 100
        x = numpy.linspace( 0, 10, nn, dtype=float )
        ym = 1.9 + 2.2 * x
        numpy.random.seed( 2345 )
        y = numpy.random.poisson( ym, size=nn )

        limits = [0,4]

        if plot :
            plt.plot( x, ym, 'k-' )
            plt.plot( x, y, 'r.' )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )

        bf = AmoebaFitter( x, model, errdis="poisson" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )
        logz0 = bf.getLogZ( limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logl  ", fmt( logl0 ) )

        errdis = PoissonErrorDistribution( )
        problem = ClassicProblem( model, xdata=x, ydata=y )
        logz1, logl1 = plotErrdis2d( errdis, problem, limits=limits, max=0,
                    plot=plot )
        if plot :
            plt.plot( pars[0], pars[1], 'k.' )

        print( "logZ  ", fmt( logz1 ), "   logL  ", fmt( logl1 ) )

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=limits[0], highLimits=limits[1] )
        ns = PhantomSampler( x, model, y, distribution='poisson', verbose=0, seed=23456 )

        logE = ns.sample()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        samples = ns.samples

        print( "pars  ", fmt( par2 ), fmt( samples.maxLikelihoodParameters ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        self.assertTrue( abs( logz2 - logz1 ) < 3 * dlz2 )

        parevo =samples.getParameterEvolution()
        llevo = samples.getLogLikelihoodEvolution()
        lwevo = samples.getLogWeightEvolution()

        assertAAE( numpy.sum( numpy.exp( lwevo ) ), 1.0 )

        if plot :
            plt.plot( parevo[:,0], parevo[:,1], 'k,' )
            plt.show()


    def test8( self ) :
        print( "====test8  Uniform ================" )
        plot = self.doplot

        nn = 200
        x = numpy.linspace( 0, 2, nn, dtype=float )
        ym = 0.3 + 5.4 * x

        y = numpy.round( ym )

        limits = [-1,2]

        model = PolynomialModel( 1 )
        model.setLimits( lowLimits=[0,0], highLimits=[10,10] )

        ns = PhantomSampler( x, model, y, distribution='uniform', verbose=1,
            engines=["chord"] )
#            engines=["galilean"] )

#        ns.engines[0].debug = True
        ns.distribution.setLimits( [0.1, 1000] )
        ns.minimumIterations = 501

        logE = ns.sample( plot=plot )
        plt.show()

        par2 = ns.parameters
        logE = ns.logZ
        dlz2 = ns.logZprecision
        logz2 = ns.logZ
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz2 ), " +- ", fmt( dlz2 ) )

        eng = ns.engines[0]
        eng.calculateUnitRange()
        print( eng.unitRange )
        print( eng.unitMin )

        for w in ns.walkers :
            pars = w.allpars
            up = eng.domain2Unit( ns.problem, pars )
#            print( pars, up )
#            plt.plot( [up[0]], [up[1]], 'k.' )
#            plt.plot( [up[2]], [up[0]], 'r.' )
#            plt.plot( [up[1]], [up[2]], 'g.' )


#        plt.show()

        if plot :
            plt.show()

    def XXXtest9( self ) :
        print( "====test9  Bernoulli ================" )
        plot = self.doplot

        nn = 61
        x = numpy.linspace( -3, 3, nn, dtype=float )
        numpy.random.seed( 124 )
        xp = x + numpy.random.randn( nn )
        y = numpy.where( xp > 0, 1, 0 )
        print( fmt( x, max=None ) )
        print( fmt( xp, max=None ) )
        print( fmt( y, max=None ) )


        model = LogisticModel( fixed={0:1} )

        model.setLimits( lowLimits=[-3,-2], highLimits=[3,2] )

        ns = PhantomSampler( x, model, y, distribution='bernoulli', verbose=2 )
#            engines=["chord"] )
#            engines=["galilean"] )

#        ns.engines[0].debug = True
        ns.minimumIterations = 501

        logE = ns.sample( )

        yfit1 = model( x )
        par2 = ns.parameters
        logz = ns.logZ
        dlz  = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz ), " +- ", fmt( dlz ) )
        print( "logL  ", fmt( ns.distribution.logLikelihood( ns.problem, par2 ) ) )

        bf = AmoebaFitter( x, model, errdis="bernoulli" )

        pars = bf.fit( y, tolerance=1e-20 )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )

#        limits = [6,4]
        logz0 = bf.getLogZ()        # limits=limits )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

        yfit2 = model( x )

        plt.plot( x, y, 'k.' )
        plt.plot( x, yfit1, 'r-' )
        plt.plot( x, yfit2, 'g-' )
        plt.show()

    def XXXtest10( self ) :
        print( "====test10  Bernoulli ================" )
        plot = self.doplot

        nn = 1000
        x = numpy.linspace( 0, 10, nn )
        y = numpy.zeros( nn )
        y[:300] = 1
        y[900:] = 1
        y[400:600] = 2

#        Tools.printclass( UniformPrior( limits=[0,1] ) )

        dtm = DecisionTreeModel( ndim=1, kdim=[0], depth=1, dynamic=False )
#        print( dtm.result( x, [0,1] ) )

#        Tools.printclass( dtm )

        problem = CategoricalProblem( model=dtm, xdata=x, ydata=y )
        print( problem.npars, problem.ncateg )
        print( problem.ydata.shape )

        ns = PhantomSampler( problem=problem, distribution='bernoulli' )
        ns.verbose = 2

#        Tools.printclass( ns )

        logE = ns.sample( )

#        yfit1 = model( x )
        par2 = ns.parameters
        logz = ns.logZ
        dlz  = ns.logZprecision
        print( "pars  ", fmt( par2 ) )
        print( "stdv  ", fmt( ns.stdevs ) )
        print( "logZ  ", fmt( logz ), " +- ", fmt( dlz ) )
        print( "logE  ", fmt( logE ) )
        print( "logL  ", fmt( ns.distribution.logLikelihood( ns.problem, par2 ) ) )

        bf = CategoricalFitter( problem )

        pars = bf.fit( )
        print( "pars  ", fmt( pars ) )
        print( "stdv  ", fmt( bf.stdevs ) )

#        limits = [6,4]
        logz0 = bf.getLogZ( pars )
        logl0 = bf.logLikelihood
        print( "logZ  ", fmt( logz0 ), "   logL  ", fmt( logl0 ) )

#        yfit2 = model( x )

#        plt.plot( x, y, 'k.' )
#        plt.plot( x, yfit1, 'r-' )
#        plt.plot( x, yfit2, 'g-' )
#        plt.show()



if __name__ == '__main__':
    unittest.main( )




import numpy as numpy
from . import Tools
from .Formatter import formatter as fmt
from .Formatter import fma

from .Engine import Engine
from .Engine import DummyPlotter

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester

class GibbsEngine( Engine ):
    """
    Move a one parameter at a time by a random amount.

    The walker is kept when the logLikelihood > lowLhood

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ) :
        """
        Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : GibbsEngine
            to be copied
        kwargs : for Engine
            "slow", "seed", "verbose"

        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )

        self.plotter = DummyPlotter()

    def copy( self ):
        """ Return copy of this.  """
        return GibbsEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "GibbsEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()              ## work on local copy

        problem = walker.problem
        fitIndex = walker.fitIndex

        perm = self.rng.permutation( fitIndex )
        if hasattr( self, "nstep" ) :
            pm = perm
            for k in range( 1, self.nstep ) :
                perm = numpy.append( perm, pm )

        ur = self.unitRange * ( 1 + 2.0 / len( self.walkers ) )

        param = walker.allpars
        self.plotter.start( param=param )

        if self.verbose > 4 :
            print( "alpar ", fma( param ), fmt( walker.logL ), fmt( lowLhood) )
            fip = param[fitIndex]
            print( "uap   ", fma( self.domain2Unit( problem, fip, fitIndex )))
            print( "fitin ", fma( fitIndex ), self.maxtrials )
            print( "unitr ", fma( self.unitRange ) )

        steps = 0
        for c in perm :

            save = param[c]
            ptry = param.copy()
            usav = self.domain2Unit( problem, save, kpar=c )
            while True :
                step = 2 * self.rng.rand() - 1.0
                if c < len( ur ) :
                    step *= ur[c]
                utry = usav + step
                if 0 < utry < 1 : break

            kk = 0
            while True :
                kk += 1
                ptry[c] = self.unit2Domain( problem, utry, kpar=c )

                Ltry = self.errdis.updateLogL( problem, ptry, parval={c : save} )

                if self.verbose > 4 :
                    print( "Gibbs ", fma(ptry), fmt(Ltry), kk, fmt( step ) )

                if Ltry >= lowLhood:
                    self.plotter.move( param, ptry, col=0, sym=0 )
                    self.reportSuccess( )
                    param = ptry

                    update = len( self.walkers ) if append else kw
                    self.setWalker( update, problem, param, Ltry, fitIndex=fitIndex )
                    steps += 1
                    break
                elif kk < self.maxtrials :
                    self.plotter.move( param, ptry, col=5, sym=4 )
                    while True :
                        utry = usav + step * ( 2 * self.rng.rand() - 1.0 )
                        if 0 < utry < 1 : break
                    self.reportReject( )
                else :
                    self.reportFailed()
                    param[c] = save
                    break

        self.plotter.stop( param=param, name="GibbsEngine" )

        return steps                        # nr of succesfull steps





import numpy as numpy
import math

from .ErrorDistribution import ErrorDistribution
from .NoiseScale import NoiseScale
from .JeffreysPrior import JeffreysPrior
from . import Tools

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester


class ScaledErrorDistribution( ErrorDistribution ):
    """
    Base class that defines methods common to error distributions with a scale.

    GaussErrorDistribution
    LaplaceErrorDistribution
    CauchyErrorDistribution
    ExponentialErrorDistribution
    UniformErrorDistribution

    Author       Do Kester.

    """
    PARNAMES = ["scale"]

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, scale=1.0, limits=None, fixed=None, copy=None ):
        """
        Default Constructor.

        Parameters
        ----------
        scale : float
            noise scale
        limits : None or list of 2 floats [low,high]
            None : no limits implying fixed scale
            low     low limit on scale (needs to be >0)
            high    high limit on scale
            when limits are set, the scale is to be fitted
        fixed : dictionary of {int:float}
            int     list if parameters to fix permanently. Default None.
            float   list of values for the fixed parameters.

        copy : ScaledErrorDistribution
            distribution to be copied.

        """
        super( ScaledErrorDistribution, self ).__init__( fixed=fixed, copy=copy )

        if copy is None :
            self.hyperpar = NoiseScale( scale=scale, limits=limits )
        else :
            self.hyperpar = copy.hyperpar                   ## TBC copy ???

        if limits is None :
            self.fixed = self.keepFixed( {0: scale} )

    def copy( self ):
        """ Return copy of this.  """
        return ScaledErrorDistribution( copy=self )

    def setLimits( self, limits ) :
        """
        Set limits for scale.

        Parameters
        ----------
        limits : [low,high]
            low : float or array_like
                low limits
            high : float or array_like
                high limits
        """
        if self.hyperpar[0].prior is None :
            self.hyperpar[0].prior = JeffreysPrior()

        super( ScaledErrorDistribution, self ).setLimits( limits )


    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        if name == "scale" and Tools.isInstance( value, float ) :
            self.hyperpar[0].hypar = value
        else :
            super( ScaledErrorDistribution, self ).__setattr__( name, value )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == "scale" :
            return self.hyperpar[0].hypar
        else :
            return super( ScaledErrorDistribution, self ).__getattr__( name )


